import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.tensorboard import SummaryWriter
from streamer.utils.logging import TBWriter, checkdir
from collections import deque
import random


class Logger():

    def __init__(self, writer_path, log_base_every, distance_mode):

        checkdir(writer_path)
        self.writer = SummaryWriter(writer_path)
        self.writer.flush()

        self.counter = 0
        self.log_base_every = log_base_every
        self.dist_mode = distance_mode
        self.layer_logs = []
        self.layer_deqs = []

        self.rep_deque = deque(maxlen=50)



    def model(self, prediction, groundtruth):
        pass

    def avg_rep(self, layer_num, representation):

        if len(self.layer_deqs) <= layer_num:
            self.layer_deqs.append(deque(maxlen=50))

        if random.random() > 0.5: self.layer_deqs[layer_num].append(representation.detach().clone())
        if len(self.layer_deqs[layer_num])<= 1: return 0.0

        embeddings = torch.cat(list(self.layer_deqs[layer_num]), dim=0)

        # needs to be fixed
        if self.dist_mode == 'distance':
            distances = torch.cdist(embeddings, embeddings, p=2)
        elif self.dist_mode == 'similarity':
            distances = torch.mm(embeddings, embeddings.t())
        else:
            raise Exception("distance mode not implemented")

        mask = torch.tril(torch.ones_like(distances), diagonal=-1).bool()
        val = distances[mask].mean()

        return val



    def layer(self, layer_num, boundary, loss, event_size, representation):
        if len(self.layer_logs) <= layer_num:
            self.layer_logs.append([
                TBWriter(self.writer, 'scalar', f'layer {layer_num}/loss {self.dist_mode}'),
                TBWriter(self.writer, 'scalar', f'layer {layer_num}/prediction length'),
                TBWriter(self.writer, 'scalar', f'layer {layer_num}/representation avg {self.dist_mode}'),]
                                   )



        self.layer_logs[layer_num][0](loss)
        if boundary: 
            self.layer_logs[layer_num][1](event_size)
            self.layer_logs[layer_num][2](self.avg_rep(layer_num, representation))

    def __del__(self):
        self.writer.close()





import torch
import importlib, os


def getLogger(args):

    """
    Import the correct logger
    """

    dataset_name = os.path.split(args.dataset.rstrip('/'))[-1]
    if dataset_name in ['epic', 'ego4d', 'kagu', 'finegym']:
        modality = "video"
    elif dataset_name in ['vox', 'timit']:
        modality = "audio"
    else:
        quit('dataset not found')

    logger_filename  = f'streamer.loggers.{modality}_logger'
    modellib = importlib.import_module(logger_filename)

    logger_class = None
    target_logger_name =  f'{modality}Logger'
    for name, cls in modellib.__dict__.items():
        if name.lower() == target_logger_name.lower():
            logger_class = cls

    if logger_class is None:
        raise NotImplementedError("In %s.py, there should be a class name that matches %s in lowercase." % (logger_filename, target_logger_name))

    return logger_class
    





import torch, math
import torch.nn as nn
import torch.nn.functional as F
from streamer.loggers.logger import Logger
from torchvision.utils import make_grid
from streamer.utils.logging import TBWriter
import matplotlib.cm as cm
import numpy as np

class VideoLogger(Logger):

    def model(self, prediction, groundtruth, attn_img, attns):
        self.counter += 1

        # checks
        if (self.log_base_every<=0) or \
           (not len(self.layer_logs)) or \
           (self.counter%self.log_base_every != 0):
               return

        # prediction logger
        if not hasattr(self, 'model_writer'):
            self.model_writer = TBWriter(self.writer, 'image', 'prediction images')

        if isinstance(prediction, torch.Tensor):
            # prediction = F.normalize(prediction, p=2, dim=1)
            imgs_list = [((prediction+1.0)/2.0).squeeze(0), ((groundtruth+1.0)/2.0).squeeze(0)]
            self.model_writer(make_grid([ *imgs_list ], nrow=2))


        # attention logger
        if not hasattr(self, 'model_attn'):
            self.model_attn = TBWriter(self.writer, 'image', 'attention images')

        if isinstance(attns, list):
            imgs_list = [((attn_img+1.0)/2.0).squeeze(0), self.visualize_attention(attns)]
            self.model_attn(make_grid([ *imgs_list ], nrow=2))


    @torch.no_grad()
    def visualize_attention(self, attention_maps):

        # Concatenate the attention maps from all blocks
        attention_maps = torch.cat(attention_maps, dim=1) # -> [1, blocks*heads, 65, 65]

        # select only the attention maps of the CLS token
        attention_maps = attention_maps[:, :, 0, 1:] # -> [1, blocks*heads, 64]

        # Then average the attention maps of the CLS token over all the heads
        attention_maps = attention_maps.mean(dim=1) # -> [1, 64]

        # Reshape the attention maps to a square
        num_patches = attention_maps.size(-1)
        size = int(math.sqrt(num_patches))
        attention_maps = attention_maps.view(-1, size, size) # -> [1, 8, 8]

        # Resize the map to the size of the image
        attention_maps = attention_maps.unsqueeze(1)
        attention_maps = F.interpolate(attention_maps, size=(128, 128), mode='bilinear', align_corners=False)
        attention_map = attention_maps.squeeze(1) # -> [1, 128, 128]

        # Normalize the attention map to range [0, 1]
        attention_map_normalized = (attention_map - attention_map.min()) / (attention_map.max() - attention_map.min())

        # Convert the PyTorch tensor to a NumPy array
        attention_map_array = attention_map_normalized.cpu().numpy()

        # Apply the "jet" colormap
        cmap_jet = cm.get_cmap('jet')
        colored_attention_map = cmap_jet(attention_map_array)

        # Convert the RGBA image to RGB
        colored_attention_map_rgb = torch.Tensor(colored_attention_map[:, :, :, :3].tolist())[0].permute(2,0,1).cuda()
        return colored_attention_map_rgb




from collections import deque
import numpy as np


class EventDemarcation():
    r"""
    Define different Demarcation strategies

    :param str dem_mode: the type of demarcation loss: ['fixed', 'accum', 'average']
    :param str dist_mode: the type of demarcation loss: ['similarity', 'distance']
    :param float threshold: error or similarity threshold value. Only used with 'accum' and 'fixed' demarcation modes
    :param int window_size: Window size for moving average. Only used for 'average' demarcation mode
    :param str modifier_type: type of change to average: ['add', 'multiply']. Only used for 'average' demarcation mode
    :param float modifier: multiplier to change the average value as threshold. Only used for 'average' demarcation mode
    """

    def __init__(self, dem_mode, dist_mode, **kwargs):
        self.dem_mode = dem_mode
        self.dist_mode = dist_mode


        if dem_mode == 'fixed':
            self.threshold = kwargs['threshold']
        elif dem_mode == 'accum':
            self.buffer = 0.0
            self.threshold = kwargs['threshold']
            assert self.dist_mode == 'distance', "Use distance measure with accum demarcation type"
        elif dem_mode == 'average':
            self.window = deque(maxlen=kwargs['window_size'])
            self.modifier = kwargs['modifier']
            self.modifier_type =kwargs['modifier_type']
        else:
            raise NotImplementedError('dem mode %s not implemented' % dem_mode)

    def __call__(self, value):
        r"""
        Detect boundary according to demarcation mode and distance mode

        :param float value: The new error value
        :returns:
            (*bool*): True if the new value in a boundary

        """

        if self.dem_mode == 'fixed':
            if self.dist_mode == 'similarity': return (value < self.threshold)
            elif self.dist_mode == 'distance': return (value > self.threshold)

        elif self.dem_mode == 'accum':
            self.buffer += value

            if self.buffer > self.threshold:
                self.buffer = 0.0
                return True
            else: return False

        elif self.dem_mode == 'average':
            if len(self.window) == 0:
                self.window.append(value)
                return False

            avg = np.average(list(self.window))

            if self.modifier_type == 'add': avg += self.modifier
            elif self.modifier_type == 'multiply': avg *= self.modifier

            boundary = value<avg if self.dist_mode == 'similarity' else value > avg
            self.window.append(value)
            return boundary

    def reset(self):
        r"""
        Reset the window for demarcation. Only affects 'accum' and 'average' demarcation modes
        """

        if hasattr(self, 'window'):
            self.window.clear()

        if hasattr(self, 'buffer'):
            self.buffer = 0.0








import torch
import torch.distributed as dist
import numpy as np
from dataclasses import dataclass


@dataclass
class StreamerOptimizerArguments():

    world_size: int = 1
    r"""Number of gpus to distribute the dataset"""

    alpha: int = 3
    r""" The reach parameter for Hierarchical Gradient Normalization"""

    max_layers: int = 3
    r"""The maximum number of layers to stack"""

    optimize_every: int = 100
    r"""Take a gradient step every this value"""

    average_every: int = 1000
    r"""Average models across gpus every this value"""

    hgn_timescale: bool = True
    r"""Allow timescale parameter in Hierarchical Gradient Normalization"""

    hgn_reach: bool = True
    r"""Allow reach parameter in Hierarchical Gradient Normalization"""

    bp_up: bool = True
    r"""Allow bottom up optimization"""

    bp_down: bool = True
    r"""Allow top down optimization"""

    @staticmethod
    def from_args(args):
        return StreamerOptimizerArguments(
                world_size=args.world_size, 
                alpha=args.alpha, 
                max_layers=args.max_layers, 
                optimize_every=args.optimize_every,
                average_every=args.average_every,
                hgn_timescale = args.hgn_timescale,
                hgn_reach = args.hgn_reach,
                bp_up = args.bp_up,
                bp_down = args.bp_down
        )



class StreamerOptimizer():
    r"""
    The optimizer used with streamer. 
    This class takes care of optimization, Gradient normalization and averaging across gpus.

    :param StreamerOptimizerArguments args: The parameters used for the Streamer optimizer
    """

    def __init__(self, model, args:StreamerOptimizerArguments):
        self.args = args
        self.model = model
        self.curr_n_layers = 0
        self.average_counter = 0
        self.step_counter = 0
        self.reset()

    def get_param_groups(self, layer_num):
        r"""
        Calculates the parameter groups and their weights.
        For example if the layer_num is 1 and we have 4 layers, then parameter groups will be [[1],[0,2],[3]] 
        and their weights will depend on alpha but typically more on the early groups (e.g., [0.8, 0.15, 0.05])

        :param int layer_num: the index of the layer
        :returns:
            * (*List(List(int))*): List of Lists dividing the layers into groups to assign different gradient multipliers to them
            * (*List(float)*): The weights assigned to the parameter groups
        """

        less = [layer_num - i for i in range(len(self.f_params)) if layer_num-i>=0]
        more = [layer_num + i for i in range(len(self.f_params)) if layer_num+i<len(self.f_params)]
        groups = []
        for i in range(max(len(less), len(more))):
            g = set()
            if i < len(less):
                g.add(less[i])
            if i < len(more):
                g.add(more[i])
            groups.append(list(g))


        weights = torch.Tensor([1.0/(self.args.alpha**i) for i in range(len(groups))])
        weights /= weights.sum()

        return groups, weights


    def get_current_params(self):
        r"""
        Gets a reference to all parameters of all layers in the streamer model
        """

        n_layers = 1
        ll = self.model.streamer
        self.f_params = [ll.f.get_params()]
        self.p_params = [ll.p.get_params()]
        self.l_params = [ll.l.get_params()]

        while ll.above != None:
            n_layers += 1
            ll = ll.above
            self.f_params.append(ll.f.get_params())
            self.p_params.append(ll.p.get_params())
            self.l_params.append(ll.l.get_params())

        return n_layers

    def check_bp(self, layer, curr_layer):
        up, down = (layer < curr_layer), (layer > curr_layer)
        if not (up or down): return True
        if up and self.args.bp_up: return True
        if down and self.args.bp_down: return True
        return False


    def update_all_layers(self, layer):

        # return if no loss to use
        if not layer.objective_ready:
            return

        layer_num = layer.layer_num
        layer_loss = layer.l.summarize_loss(layer.main_objective)
        layer_timescale = float(layer.x_size) if self.args.hgn_timescale else 1.0

        self.p_counter[layer_num] += layer_timescale
        groups, reaches = self.get_param_groups(layer_num)

        for g_i, (group, reach) in enumerate(zip(groups, reaches)):

            # if not self.check_bp(layer_num, i): continue # need to be fixed

            # get parameters to accumulate
            params = [*self.p_params[layer_num]]
            for g_i in group: params.extend(self.f_params[g_i])

            retain_graph = g_i!=(len(groups)-1) or True
            (layer_loss*layer_timescale*reach).backward(inputs=params, retain_graph=retain_graph)
            for i in group: self.f_counter[i] += layer_timescale

        # reset loss
        self.step_counter[layer_num] += 1
        layer.objective_ready = False


    def get_gradients(self):
        r"""
        accumulates gradient on all layers' parameters from all losses in the model
        """

        ll = self.model.streamer
        self.update_all_layers(ll)

        while ll.above != None:
            ll = ll.above
            self.update_all_layers(ll)


    def equal_layers(self):
        r"""
        Determines if the streamer model has the same number of layers on all gpus

        :returns:
            * (*bool*): True if equal layers on all gpus
            * (*int*): The number of layers on the current gpu
        """
        l_global = torch.tensor([0.0]).cuda()
        l_global = self.model.streamer.get_num_layers(l_global)
        l_local = l_global.clone()

        if self.args.world_size == 1:
            return True, int(l_local.item())

        dist.all_reduce(l_global)
        return l_local*self.args.world_size == l_global, int(l_local.item())


    def reset(self):
        r"""
        Resets the counters of the optimizer
        """

        self.f_counter = {i:0 for i in range(self.args.max_layers)}
        self.p_counter = {i:0 for i in range(self.args.max_layers)}
        self.l_counter = {i:0 for i in range(self.args.max_layers)}

        self.step_counter = {i:0 for i in range(self.args.max_layers)}

    def scale_gradients(self):
        r"""
        Scales the gradients of all modules by the counters to normalize the gradients
        """

        def scale_module(params, counter):
            for p_ix, param in enumerate(params):
                if counter[p_ix] == 0: continue
                for p in param:
                    if p.grad != None: p.grad /= counter[p_ix]

        scale_module(self.f_params, self.f_counter)
        scale_module(self.p_params, self.p_counter)



    def average_models(self):
        r"""
        Average the model parameters across all gpus every :py:meth:`~StreamerOptimizerArguments.average_every`
        """

        for name, param in self.model.named_parameters():
            dist.all_reduce(param.data, op=dist.ReduceOp.SUM)
            param.data /= self.args.world_size



    def step(self):
        r"""
        Call the optimizer, which calulates the gradients and accumulates it on the parameters.
        This function does not actually do gradient stepping. It has a counter that does it every :py:meth:`~StreamerOptimizerArguments.optimize_every`
        """

        # === AVERAGE MODELS === #
        if self.args.world_size > 1: self.average_counter += 1

        equal, n_layers = self.equal_layers()
        if n_layers > self.curr_n_layers: 
            self.curr_n_layers = self.get_current_params()

        if equal and self.average_counter >= self.args.average_every:
            self.average_models()
            self.average_counter = 0


        # === OPTIMIZATION === #
        self.get_gradients() # backwards
        if self.step_counter[self.curr_n_layers-1] >= self.args.optimize_every:
            self.scale_gradients()
            self.model.optimize_model()
            self.reset()








import argparse

def bool_flag(s):
    r"""
    Parse boolean arguments from the command line.

    :param str s: command line string argument
    :returns bool: the boolean value of the argument
    """
    FALSY_STRINGS = {"off", "false", "0"}
    TRUTHY_STRINGS = {"on", "true", "1"}
    if s.lower() in FALSY_STRINGS:
        return False
    elif s.lower() in TRUTHY_STRINGS:
        return True
    else:
        raise argparse.ArgumentTypeError("invalid value for a boolean flag")


def parse_args():
    r"""
    parser function for cli arguments
    """

    parser = argparse.ArgumentParser(description='Streamer parser')

    # basic arguments
    parser.add_argument('--dataset', type=str, default='data/epic', help='Path to dataset folder')
    parser.add_argument('--output', type=str, default='out', help='Path to output folder')
    parser.add_argument('--type', type=str, choices=['streamer', 'dreamer'], default='streamer', help='model type')
    parser.add_argument('--name', type=str, default='my_model', help='name of the experiment')

    # Platform settings
    parser.add_argument('--p_name', type=str, default='job', help='platform job name for slurm')
    parser.add_argument('--p_device', type=str, choices=['gpu', 'slurm', 'cpu', 'mps'], default='gpu', help='platform device')
    parser.add_argument('--p_partition', type=str, default='general', help='platform partition with slurm')
    parser.add_argument('--p_n_nodes', type=int, default=1, help='platform number of nodes with slurm')
    parser.add_argument('--p_n_gpus', type=int, default=1, help='platform number of gpus')
    parser.add_argument('--p_n_cpus', type=int, default=2, help='platform number of total cpus per process')
    parser.add_argument('--p_ram', type=int, default=10, help='platform number of total ram in GB')
    parser.add_argument('--p_backend', type=str, choices=['nccl', 'gloo'], default='nccl', help='platform backend')
    parser.add_argument('--p_verbose', type=bool_flag, default=True, help='platform verbose')
    parser.add_argument('--p_logs', type=str, default='./logs', help='platform console logs')

    # dataset arguments
    parser.add_argument('--frame_size', type=int, nargs='+', default=[128,128], help='frame size')
    parser.add_argument('--num_workers', type=int, default=1, help='number of workers for dataloader')
    parser.add_argument('--dataset_split', type=str, default='train', choices=['train','test'], help='dataset split')
    parser.add_argument('--dataset_percent', type=int, default='100', help='dataset percentage to use')


    # model arguments
    parser.add_argument('--feature_dim', type=int, default=1024, help='feature dimension of each representation')
    parser.add_argument('--max_layers', type=int, default=3, help='maximum number of hierarchical layers')
    parser.add_argument('--evolve_every', type=int, default=1e5, help='add new layer every x steps')
    parser.add_argument('--init_layers', '-il', type=int, default=1, help='number of layers to initialize the model')
    parser.add_argument('--init_ckpt', '-ic', type=str, default='', help='ckpt path to load, overrides init_layers with layers count')
    parser.add_argument('--buffer_size', '-bs', type=int, default=10, help='buffer size for f function')
    parser.add_argument('--force_fixed_buffer', '-fb', type=bool_flag, default=False, help='force fixed buffer size')
    parser.add_argument('--l1_loss_weight', '-l1', type=float, default=100.0, help='l1 loss weight for dreamer')
    parser.add_argument('--demarcation_mode', type=str, default='average', choices=['fixed', 'accum', 'average'], help='Demarcation type')
    parser.add_argument('--distance_mode', type=str, default='similarity', choices=['similarity', 'distance'], help='distance type')
    parser.add_argument('--force_base_dist', type=bool_flag, default=False, help='force base layer distance mode to be distance')
    parser.add_argument('--normalize_imgs', type=bool_flag, default=False, help='normalize input images along channel dimension')
    parser.add_argument('--window_size', type=int, default=50, help='window size for average demarcation type')
    parser.add_argument('--modifier_type', type=str, default= 'multiply', choices=['add', 'multiply'], help='modifier type for average demarcation type')
    parser.add_argument('--modifier', type=float, default=1.0, help='modifier for average demarcation type')
    parser.add_argument('--loss_threshold', '-lt', type=float, default=0.1, help='loss threshold')

    # training arguments
    parser.add_argument('--lr', type=float, default=1e-4, help='learning rate')
    parser.add_argument('--alpha', type=float, default=3.0, help='reach hyperparameter for normalization')
    parser.add_argument('--optimize_every', type=int, default=100, help='optimize model every x steps of highest layer')
    parser.add_argument('--average_every', type=int, default=1000, help='average weights of models every x')
    parser.add_argument('--optimize', type=bool_flag, default=True, help='optimize the model')
    parser.add_argument('--save_every', type=int, default=25000, help='save model every x steps')
    parser.add_argument('--hgn_timescale', type=bool_flag, default=True, help='optimize higher layers with higher learning rate')
    parser.add_argument('--hgn_reach', type=bool_flag, default=True, help='optimize with reach of influence')
    parser.add_argument('--bp_up', type=bool_flag, default=True, help='backpropagate upwards to higher layers')
    parser.add_argument('--bp_down', type=bool_flag, default=True, help='backpropagate downwards to lower layers')
    parser.add_argument('--fp_up', type=bool_flag, default=True, help='forward propagate upwards to higher layers')
    parser.add_argument('--fp_down', type=bool_flag, default=True, help='forward propagate downwards to lower layers')

    # advanced arguments
    parser.add_argument('--dbg', action='store_true', help='Flag for debugging and development')
    parser.add_argument('--tb', action='store_true', help='Flag for tb logging')
    parser.add_argument('--log_prefix', type=str, default='/data/D2/datasets/epic_kitchen/videos/', help='prefix to write in json file')
    parser.add_argument('--log_postfix', type=str, default='MP4', help='postfix extension to write in json file')
    parser.add_argument('--log_base_every', type=int, default='1000', help='log base signal to tensorboard every')

    return parser.parse_args()






import importlib

def getOptimizer(args, model):

    model_filename  = f'streamer.optimizers.{args.type}_optimizer'
    modellib = importlib.import_module(model_filename)

    optimizer_class = None
    optimizer_args_class = None
    target_optimizer_name =  f'{args.type}Optimizer'
    target_optimizer_args_name =  f'{args.type}OptimizerArguments'
    for name, cls in modellib.__dict__.items():
        if name.lower() == target_optimizer_name.lower():
            optimizer_class = cls
        elif name.lower() == target_optimizer_args_name.lower():
            optimizer_args_class = cls

    if optimizer_class is None:
        raise NotImplementedError("In %s.py, there should be a class name that matches %s in lowercase." % (model_filename, target_optimizer_name))

    if optimizer_args_class is None:
        raise NotImplementedError("In %s.py, there should be a class name that matches %s in lowercase." % (model_filename, target_optimizer_args_name))

    optimizer_args = optimizer_args_class.from_args(args)
    optimizer = optimizer_class(model, optimizer_args)
    return optimizer




import importlib, os
from torch.utils.data import DataLoader

def find_dataset_using_name(args):
    """Import the module "data/[dataset_name]_dataset.py".

    In the file, the class called DatasetNameDataset() will
    be instantiated. It has to be a subclass of BaseDataset,
    and it is case-insensitive.
    """

    dataset_name = os.path.split(args.dataset.rstrip('/'))[-1]
    dataset_filename = "streamer.datasets." + dataset_name + "_dataset"
    datasetlib = importlib.import_module(dataset_filename)

    dataset_class = None
    dataset_args_class = None
    target_dataset_name = dataset_name.replace('_', '') + 'dataset'
    target_dataset_args_name = dataset_name.replace('_', '') + 'datasetarguments'
    for name, cls in datasetlib.__dict__.items():
        if name.lower() == target_dataset_name.lower():
            dataset_class = cls
        elif name.lower() == target_dataset_args_name.lower():
            dataset_args_class = cls

    if dataset_class is None:
        raise NotImplementedError("In %s.py, there should be a class name that matches %s in lowercase." % (dataset_filename, target_dataset_name))

    if dataset_args_class is None:
        raise NotImplementedError("In %s.py, there should be an argument class name that matches %s in lowercase." % (dataset_filename, target_dataset_args_name))


    dataset_args = dataset_args_class.from_args(args)
    dataset = dataset_class(dataset_args)
    args.snippet_size = dataset.snippet_size
    loader = DataLoader(dataset=dataset, batch_size=1, num_workers= args.num_workers, pin_memory=True)

    return loader



import torch
import numpy as np, os, random
import torch.distributed as dist
import torch.backends.cudnn as cudnn
from streamer.utils.logging import checkdir
import streamer.loggers as loggers


def setup_for_distributed(is_master):
    import builtins as __builtin__
    builtin_print = __builtin__.print

    def print(*args, **kwargs):
        force = kwargs.pop('force', False)
        if is_master or force:
            builtin_print(*args, **kwargs)

    __builtin__.print = print

def fix_random_seeds(seed=31):
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)


def init_gpu(global_rank, local_rank, args):

    # Initialize distributed environment
    args.local_rank = local_rank
    args.global_rank = global_rank
    args.device = torch.device("cuda:{}".format(args.local_rank))
    torch.cuda.set_device(args.device)

    fix_random_seeds()
    cudnn.benchmark = True

    args.main = (args.global_rank == 0)
    if args.p_device != 'slurm': setup_for_distributed(args.main)

    # Tensorboard logger
    args.logger = None
    if args.tb:
        logger_class = loggers.getLogger(args)
        logger_path = os.path.join(args.exp_output, 'logs', str(args.global_rank))
        args.logger = logger_class(logger_path, args.log_base_every, args.distance_mode)

    args.json_dir = checkdir(f'{args.exp_output}/jsons/{args.global_rank}')






import numpy as np


class MemBuffer():
    r"""
    Buffer class that maintains size of inputs to the temporal encoding model.
    The buffer replaces low prediction error inputs with high prediction error values
    while maintaining the order they were received in.

    :param int buffer_size: The maximum buffer size to maintain
    :param str dist_mode: The distance mode for error calculation (e.g., 'similarity', 'distance')

    """
    def __init__(self, buffer_size=20, dist_mode='similarity'):
        r"""
        init function of the MemBuffer Class
        """


        self.dist_mode = dist_mode
        self.buffer_size = buffer_size
        self.reset_buffer()

    def __len__(self):
        r"""
        Calculates the current length of the buffer

        :returns:
            (*int*): The size of the buffer

        """
        return len(self.inputs)

    def __bool__(self):
        r"""
        Evaluates if the buffer has at least one input

        :returns:
            (*bool*): True is the buffer has one or more inputs
        """
        return len(self.inputs) > 0

    def __repr__(self):
        r"""
        Returns a representation string of the buffer

        :returns:
            (*str*): String describing the number of inputs and error range
        """
        if not self:
            return f'Buffer empty'

        return f'Buffer has {len(self)} values with errors from {min(self.error)} to {max(self.error)}'

    def add_input(self, x, err):
        r"""
        Adds a single input and its corresponding error value to the buffer.
        If the buffer is full, this function takes care of replacing the low prediction error value if the new value is higher.

        :param torch.Tensor x: The input tensor to be added to the buffer
        :param torch.Tensor err: The corresponding error or similarity value.
        """

        self.counter += 1

        # buffer not full
        if len(self.inputs) < self.buffer_size:
            self.inputs.append(x)
            self.error.append(err)
            self.pos.append(len(self.inputs)-1)

        # buffer full
        else:

            if self.dist_mode == 'similarity':
                index = self.error.index(max(self.error))
                if err < self.error[index]:
                    self.replace(index, x, err)

            elif self.dist_mode == 'distance':

                index = self.error.index(min(self.error))
                if err > self.error[index]:
                    self.replace(index, x, err)

    def replace(self, index, x, err):
        r"""
        Replaces an index in the buffer with a new input and its corresponding error

        :param int index: The index of the input to be replaced
        :param torch.Tensor x: The input tensor to be added to the buffer
        :param torch.Tensor err: The corresponding error or similarity value.
        """
        self.inputs[index] = x
        self.error[index] = err

        pos_arr = np.array(self.pos)
        pos_arr[np.array(self.pos) > self.pos[index]] -= 1 
        pos_arr[index] = self.buffer_size -1
        self.pos = pos_arr.tolist().copy()


    def get_inputs(self):
        r"""
        Returns the current inputs in the buffer in the correct order they were received.
        
        :returns:
            (*List(torch.Tensor)*): A list of tensors in the buffer

        """
        if not self:
            return []

        return [self.inputs[self.pos.index(i)] for i in range(len(self.pos))]

    def reset_buffer(self):
        r"""
        Resets the buffer. Can be called after a boundary has been detected
        """
        self.counter = 0
        self.inputs = []
        self.pos = []
        self.error = []






import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision.io import VideoReader
import os, os.path as osp, importlib, cv2, json, copy
from streamer.utils.logging import JsonLogger, VideoWriterInference
from torchvision import transforms as T
from tqdm import tqdm
import streamer.models.networks as networks
from streamer.models.model import StreamerModel, StreamerModelArguments
from streamer.optimizers.streamer_optimizer import StreamerOptimizer, StreamerOptimizerArguments
from time import sleep
from torchvision.utils import make_grid
import streamer.optimizers

class InferenceLoader():
    r"""
    Dataloader used by InferenceModel to generate the images from the video

    :param str modality: only video is supported now
    :param str filename: Video filename to load
    :param float snippet_size: How many seconds per frame (typically 0.5)
    """

    def __init__(self, modality, filename, snippet_size):
        if modality =='video':
            self.transform = T.Compose([
                        T.Resize([128, 128], antialias=True),
                    ])
            self.filename = filename
            self.snippet_size = snippet_size

        elif modality == 'audio':
            raise NotImplementedError('Audio not implemented yet')

        self.init_file(self.filename)
    
    def init_file(self, file):
        self.reader = VideoReader(self.filename, 'video')
        self.metadata = self.get_metadata(self.reader)
        self.frame_interval = int(float(self.metadata['video']['fps'][0]) * self.snippet_size)
        self.num_frames = int(self.metadata['video']['fps'][0]*self.metadata['video']['duration'][0]//self.frame_interval)+1


    # fixed version from pytorch VideoReader
    def get_metadata(self, reader):
        """Returns video metadata

        Returns:
            (dict): dictionary containing duration and frame rate for every stream
        """
        if reader.backend == "pyav":
            metadata = {}  # type:  Dict[str, Any]

            for stream in reader.container.streams:
                if stream.type not in metadata:
                    if stream.type == "video":
                        rate_n = "fps"
                    else:
                        rate_n = "framerate"
                    metadata[stream.type] = {rate_n: [], "duration": []}

                rate = stream.average_rate if stream.average_rate is not None else stream.sample_rate

                try:
                    metadata[stream.type]["duration"].append(float(stream.duration * stream.time_base))
                    metadata[stream.type][rate_n].append(float(rate))
                except:
                    pass
            return metadata
        return reader._c.get_metadata()


    def __iter__(self):
        
        for frame_ix, frame in enumerate(self.reader):

            if frame_ix%self.frame_interval==0:
                frame_tens = (self.transform(frame['data']/255.0)*2.0 -1.0).unsqueeze(0)
                yield frame_tens



class InferenceModel():
    r"""
    Inference model used to run inference on a video using pretrained weights

    :param dict checkpoint: the loaded checkpoint of pretrained model
    :param Logger logger: tensorboard logger, if needed

    """

    def __init__(self, checkpoint, logger=None):


        self.checkpoint = checkpoint
        ckpt = torch.load(checkpoint)
        self.modality = ckpt['model_modality']
        self.snippet_size = ckpt['model_snippet_size']
        self.args = ckpt['model_args']
        self.logger = logger
        self.jsonlogger = JsonLogger(is_inference=True, snippet_size=self.snippet_size)

        self.model = self.get_model()

    def get_model(self):
        r"""
        Creates a streamer model and initializes it with ckpt.

        :returns:
            (:py:class:`~streamer.models.model.StreamerModel`): The initialized StreamerModel and moved to CUDA device

        """

        arguments = StreamerModelArguments(**self.args)
        arguments.init_ckpt = self.checkpoint

        if self.modality == 'video':
            model = StreamerModel(arguments, logger=self.logger, encoder=networks.CNNEncoder, decoder=networks.CNNDecoder)
        elif self.modality == 'audio':
            model = StreamerModel(arguments, logger=self.logger, encoder=None, decoder=None)

        model.modality = self.modality
        return model.cuda()


    def get_optimizer(self):
        r"""
        Creates an optimizer. Not needed in this implementation.

        :returns:
            (:py:class:`~streamer.optimizers.streamer_optimizer.StreamerOptimizer`): The optimizer to be used.

        """

        arguments = StreamerOptimizerArguments()
        optimizer = StreamerOptimizer(self.model, arguments)
        return optimizer


    def __call__(self, filename):
        r"""
        The function to run inference on a video and save logs.

        :param str filename: video path to run inference on.
        :returns:
            (*dict*): boundary results for all layers in a dictionary format
        """

        folder_path = osp.join(log_path,osp.splitext(osp.basename(filename))[0])
        os.makedirs(folder_path, exist_ok=True)

        loader = InferenceLoader(self.modality, filename, self.snippet_size)
        for frame in tqdm(loader, total=loader.num_frames):
            self.model(frame.cuda())

        hierarchy = self.model.extract_rep()
        json_file = self.jsonlogger(filename, loader.metadata['video']['duration'][0], hierarchy)

        # reset and close writer
        self.model.reset_model()
        del video_writer

        return json_file




from streamer.loggers.logger import Logger
import torch

class AudioLogger(Logger):


    def model(self, prediction, groundtruth):
        self.counter += 1

        # checks
        if (self.log_base_every<=0) or \
           (not len(self.layer_logs)) or \
           (self.counter%self.log_base_every != 0):
               return

        if isinstance(prediction, torch.Tensor):
            pass






import torch
import importlib, os
from streamer.models.networks import CNNEncoder, CNNDecoder
from streamer.models.model import StreamerModel, StreamerModelArguments
from streamer.models.inference_model import InferenceModel


def getModel(args):
    """
    Import the correct model
    """

    dataset_name = os.path.split(args.dataset.rstrip('/'))[-1]
    if dataset_name in ['epic', 'ego4d', 'kagu', 'finegym']:
        modality = "video"
    elif dataset_name in ['vox', 'timit']:
        modality = "audio"
    else:
        quit('dataset not found')

    arguments = StreamerModelArguments.from_args(args)

    if modality == 'video':
        model = StreamerModel(arguments, logger=args.logger, encoder=CNNEncoder, decoder=CNNDecoder)
    elif modality == 'audio':
        model = StreamerModel(arguments, logger=args.logger, encoder=None, decoder=None)

    model.modality = modality
    return model
    



import torch
import torch.nn as nn
import torch.nn.functional as F
from streamer.utils.events import CounterDetector
from streamer.utils.events import Patcher
from torch.utils.tensorboard import SummaryWriter
from streamer.utils.logging import TBWriter
from collections import deque
import streamer.models.networks as networks
from dataclasses import dataclass
from streamer.core.demarcation import EventDemarcation
from streamer.core.loss import StreamerLoss
from streamer.core.buffer import MemBuffer
import types

@dataclass
class StreamerLayerArguments():

    # architecture arguments
    max_layers: int
    r"""The maximum number of layers to stack"""

    feature_dim: int
    r"""Feature dimension of the model embeddings"""

    evolve_every: int
    r"""Create/stack a new layer every 'evolve_every' """

    buffer_size: int
    r"""Maximum input buffer size to be used"""

    loss_threshold: float
    r"""Loss threshold value. Not used in average demarcation mode"""

    lr: float
    r"""Learning rate to be used in all modules"""

    reps_fn: types.FunctionType 
    r"""Function to aggregate representations from all layers"""

    snippet_size: float
    r"""Snippet size of input video (seconds/image). Typically 0.5 seconds per image"""

    demarcation_mode: str = 'average'
    r"""Demarcation mode used to detect boundaries"""

    distance_mode: str = 'similarity'
    r"""Distance mode for loss calculation"""

    force_base_dist: bool = False
    r"""Force the lowest layer to use MSE instead of Cosine Similarity"""

    window_size: int = 50
    r"""Window size for average demarcation mode"""

    modifier_type: str = 'multiply'
    r"""Modifier type to apply to average demarcation mode ['multiply', 'add']"""

    modifier: float = 1.0
    r"""Modifier to apply to average demarcation mode"""

    force_fixed_buffer: bool = False
    r"""Force the buffer to be fixed (not replacing inputs) by triggering a boundary when buffer is full"""

    l1_loss_weight: float = 1.0
    r"""L1 loss weight. Not used in the current loss implementation"""

    @staticmethod
    def from_model_args(args):
        return StreamerLayerArguments(
                max_layers=args.max_layers,
                feature_dim=args.feature_dim,
                evolve_every=args.evolve_every,
                buffer_size=args.buffer_size,
                force_fixed_buffer=args.force_fixed_buffer,
                loss_threshold=args.loss_threshold,
                lr=args.lr,
                reps_fn=args.reps_fn,
                snippet_size=args.snippet_size,
                demarcation_mode=args.demarcation_mode,
                distance_mode=args.distance_mode,
                force_base_dist=args.force_base_dist,
                window_size=args.window_size,
                modifier_type=args.modifier_type,
                modifier=args.modifier,
                l1_loss_weight=args.l1_loss_weight,
        )

class StreamerLayer(nn.Module):
    r"""
    STREAMER layer implementation. 

    This layer can:
        * Create/stack other :py:class:`StreamerLayer` layers recursively for a maximum of :py:class:`~StreamerLayerArguments.max_layers`
        * Call the other :py:class:`StreamerLayer` layers by propagating current representation
        * Calculate and store the loss for the :py:class:`~streamer.optimizers.streamer_optimizer.StreamerOptimizer` to use it

    :param StreamerLayerArguments args: arguments provided to every streamer layer
    :param int layer_num: the index of the current layer in the layers stack
    :param int init_count: used to create more layers at initialization. Useful for Inference model using pretrained weights.
    :param torch.nn.Module encoder: Encoder class to be used at this layer. Passed later to the :py:class:`~streamer.models.networks.TemporalEncoding` module
    :param torch.nn.Module decoder: Decoder class to be used at this layer. Passed later to the :py:class:`~streamer.models.networks.HierarchicalPrediction` module
    :param Logger logger: Logger to be used for tensorboard.
    """

    def __init__(self, 
                 args, 
                 layer_num, 
                 init_count,
                 encoder=None, 
                 decoder=None,
                 logger=None
            ):

        super(StreamerLayer, self).__init__()

        # === References and Counters === #
        self.args = args
        self.logger = logger
        self.layer_num = layer_num
        self.init_count = init_count
        self.distance_mode = 'distance' if (args.force_base_dist and self.layer_num == 0) else args.distance_mode
        # self.global_distance_mode = args.distance_mode
        self.preprocess = (encoder != None)
        self.above = None
        self.layer_creator = CounterDetector(count=self.args.evolve_every)

        # === Memory Buffer === #
        self.buffer = MemBuffer(args.buffer_size, self.distance_mode)

        # === F function === #
        self.f = networks.TemporalEncoding(
                                feature_dim=args.feature_dim,
                                buffer_size=args.buffer_size,
                                lr=args.lr,
                                num_layers=2,
                                n_heads=8,
                                encoder=encoder,
                                patch=False,
                                )

        # === P function === #
        self.p = networks.HierarchicalPrediction(
                            feature_dim=args.feature_dim,
                            max_layers=args.max_layers,
                            lr=args.lr,
                            layer_num=self.layer_num,
                            num_layers=2,
                            n_heads=8,
                            decoder=decoder,
                            patch=False,
                            )

        # === Loss Function === #
        self.l = StreamerLoss(dist_mode = self.distance_mode)

        # === Demarcation Function === #
        self.demarcation = EventDemarcation(
                dem_mode = args.demarcation_mode,
                dist_mode = self.distance_mode,
                threshold = args.loss_threshold,
                window_size = args.window_size,
                modifier = args.modifier,
                modifier_type = args.modifier_type
                )



        # reset and clear
        self.reset_layer()
        print(f'created layer {self.layer_num}')

        # create more parent layers at init if needed
        self.create_parent(self.init_count>1)

    def reset_layer(self):
        r"""
        Reset function to be used at the beginning of a new video.
        Recursively applied to every layer.
        """

        self.buffer.reset_buffer()
        self.demarcation.reset()
        self.hierarchy_boundaries = [0]
        self.hierarchy_attn = []
        self.representation = None
        self.attn_img = None
        self.attns = None
        self.context = None
        self.objective_ready = False
        self.last_base_counter = -1
        if self.above != None: self.above.reset_layer()


    def extract_rep(self, hierarchy):
        r"""
        Extracts the hierarchy for json logging.
        Recursively applied to every layer.

        :param dict hierarchy: hierarchy dictionary to be filled with boundaries
        """

        hierarchy['boundaries'][self.layer_num] = self.hierarchy_boundaries.copy()
        if self.above != None: 
            self.above.extract_rep(hierarchy)

    def optimize_layer(self):
        r"""
        Optimization step function. Steps then zeros the gradients. 
        Calls the `step_params()` and `zero_grad()` functions of every module. (e.g., :py:meth:`~streamer.models.networks.TemporalEncoding.step_params`)
        Recursively applied to every layer.
        """

        # Step the optimizers
        self.f.step_params()
        self.p.step_params()
        self.l.step_params()

        # zero the gradients
        self.f.zero_grad()
        self.p.zero_grad()
        self.l.zero_grad()

        # reset representations because parameters have changed
        self.representation = None
        self.context = None

        if self.above != None:
            self.above.optimize_layer()

    def get_num_layers(self, num):
        r"""
        Recursive function to get the total number of layers

        :param int num: current number of layers at the previous layer
        :returns:
            (*int*): Previous num of layers + 1
        """
        if self.above != None:
            return self.above.get_num_layers(num+1)
        else: return num+1

    def create_parent(self, create):
        r"""
        Function to create/stack another :py:class:`StreamerLayer` layer

        :param bool create: only add another layer if create is True
        """
        if (not self.above) and (create) and (self.layer_num<(self.args.max_layers-1)) and (self.training):
            self.above = StreamerLayer(self.args, 
                                       self.layer_num+1, 
                                       self.init_count-1, 
                                       logger=self.logger).cuda()


    def predict(self):
        r"""
        Prediction function that calls the :py:class:`~streamer.models.networks.TemporalEncoding` and :py:class:`~streamer.models.networks.HierarchicalPrediction` modules
        """

        # === prepare input === #
        f_in = torch.cat(self.buffer.get_inputs(), dim=0)
        self.attn_img = f_in[0]

        # === Temporal Encoding === #
        '''
        input: [S, 3, 128, 128] or [S, 1024]
        outputs: [1, 1024]
        '''
        self.representation, self.attns = self.f(f_in)
        if self.distance_mode=="similarity": self.representation = F.normalize(self.representation, p=2, dim=-1)

        # === Hierarchical Prediction === #
        '''
        input: [L, 1024]
        outputs: [1, 3, 128, 128] or [1, 1024]
        '''
        reps = self.args.reps_fn(self.layer_num)
        self.context = self.p(reps)


    def forward(self, x, base_counter):
        r"""
        Forward propagation function for a layer. Recursively calls the layer above at event boundary determined by the :py:class:`~streamer.core.demarcation.EventDemarcation` module.

        :param torch.Tensor x: the input feature vector [1, feature_dim] or image [1, 3, H, W]
        :param int base_counter: the location of this input in the video for timescale caluation in the :py:class:`~streamer.optimizers.streamer_optimizer.StreamerOptimizer`
        """

        # create parent if needed
        self.create_parent(self.layer_creator())
        self.x_size = base_counter - self.last_base_counter
        self.last_base_counter = base_counter

        # check if layer is new or recently optimized
        if self.context == None:
            self.buffer.reset_buffer()
            self.buffer.add_input(x, 0.0)
            self.predict()
            return 

        # check loss of prediction with input
        self.main_objective, demarcation_signal = self.l(self.context, x)
        self.objective_ready = True

        # check boundary
        boundary_demarcation = self.demarcation(demarcation_signal.item())
        boundary_buffer =  self.args.force_fixed_buffer and self.buffer.counter >= self.buffer.buffer_size
        boundary = boundary_demarcation or boundary_buffer

        # Tensorboard logging
        if self.logger != None: 
            self.logger.layer(self.layer_num, 
                              boundary,
                              demarcation_signal.item(),
                              self.buffer.counter,
                              self.representation.detach())

        # new event created at this level, send current event upwards
        if boundary:

            # save boundary values for json
            self.hierarchy_boundaries.append(base_counter*self.args.snippet_size)

            # if above layer exists, send it up
            if self.above != None: self.above(self.representation.detach().clone(), base_counter=base_counter)
            self.buffer.reset_buffer()


        # represent and predict
        self.buffer.add_input(x, demarcation_signal.item())
        self.predict()




import torch
from torch.utils.tensorboard import SummaryWriter
import os, shutil, random, json, copy, cv2
import os.path as osp
import numpy as np
from skvideo.io import FFmpegWriter
from PIL import Image, ImageDraw, ImageFont
from collections import deque

def checkdir(path, reset = True):
    if os.path.exists(path):
        if reset:
            shutil.rmtree(path)
            os.makedirs(path)
    else:
        os.makedirs(path)

    return path


def setup_output(args):

    assert os.path.exists(args.dataset)
    assert (args.dbg) or (not os.path.exists(os.path.join(args.output, args.name))), 'experiment logs exist'

    # create output directory
    args.exp_output = checkdir(os.path.join(args.output, args.name))
    args.ckpt_dir = checkdir(f'{args.exp_output}/checkpoints')
    args.world_size = args.p_n_gpus * args.p_n_nodes

    # save commandline arguments and logs
    json.dump(args.__dict__, open(osp.join(args.exp_output, 'args.json'), 'w'), indent=2)
    args.p_logs = osp.join(args.exp_output, 'console_logs')

    return args


class TBWriter(object):

    def __init__(self, writer, data_type, tag, mul = 1, add = 0, fps = 4):

        self.step = 0
        self.mul = mul
        self.add = add
        self.fps = fps

        self.writer = writer
        self.type = data_type
        self.tag = tag

    def __call__(self, data, step = None, flush = False, metadata=None, label_img=None):

        counter = step if step != None else self.step*self.mul+self.add

        if self.type == 'scalar':
            self.writer.add_scalar(self.tag, data, global_step = counter)
        elif self.type == 'scalars':
            self.writer.add_scalars(self.tag, data, global_step = counter)
        elif self.type == 'image':
            self.writer.add_image(self.tag, data, global_step = counter)
        elif self.type == 'video':
            self.writer.add_video(self.tag, data, global_step = counter, fps = self.fps)
        elif self.type == 'figure':
            self.writer.add_figure(self.tag, data, global_step = counter)
        elif self.type == 'text':
            self.writer.add_text(self.tag, data, global_step = counter)
        elif self.type == 'histogram':
            self.writer.add_histogram(self.tag, data, global_step = counter)
        elif self.type == 'embedding':
            self.writer.add_embedding(mat=data, metadata=metadata, label_img=label_img, global_step=counter, tag=self.tag)

        self.step += 1

        if flush:
            self.writer.flush()


class JsonLogger():

    def __init__(self, 
                 is_inference = False,
                 snippet_size = 0.5,
                 prefix = '',
                 postfix = '',
                 json_dir =''
                 ):

        # arguments
        self.is_inference = is_inference
        self.snippet_size = snippet_size
        self.prefix = prefix
        self.postfix = postfix
        self.json_dir = json_dir


        self.json_template = dict(file="filename",
                             fileType = "video/mp4",
                             cursor = 0,
                             duration=100,
                             zoom=1,
                             layers=[])

        self.layer_template = dict(name = "layer 0",
                                   order = 0,
                                   annots = []
                                   )

        self.annot_template = dict(start=0.0,
                                   end=0.5,
                                   action="N/A",
                                   colour =f'rgb{0,0,0}',
                                   representation="null")

    def get_random_colors(self):
        result = []
        ind = random.randint(0, 2)
        for col in range(3):
            if col==ind: res = random.randint(75,100)
            else: res = random.randint(130, 250)
            result.append(res)

        return f'rgb({result[0]}, {result[1]}, {result[2]})'

    def get_colors(self, attn):
        if attn == None or attn == -1.0:
            return self.get_random_colors()
        
        result = (np.array([255, 255, 255]) * attn).astype(np.uint8).tolist()
        return f'rgb({result[0]}, {result[1]}, {result[2]})'

    def get_attn_norm(self, l):
        attn_list = [item/max(sublist) for sublist in l for item in sublist]
        return attn_list

    def create_layer_with_attn(self, boundaries, layer_num,  attention="null", representation="null"):

        # for edge cases where a new layer starts in the middle of processing a video
        if isinstance(attention, list) and len(attention)+1 < len(boundaries):
            attention = [-1.0 for  _ in range(len(boundaries)-len(attention)-1)] + attention

        if isinstance(representation, list) and len(representation)+1 < len(boundaries):
            representation = ["null" for  _ in range(len(boundaries)-len(representation)-1)] + representation


        layer_template = copy.deepcopy(self.layer_template)
        layer_template["name"] = f'layer {layer_num}'
        layer_template["order"] = layer_num

        for i in range(len(boundaries)-1):

            annot_template = copy.deepcopy(self.annot_template)
            annot_template["start"] = boundaries[i]
            annot_template["end"] = boundaries[i+1]
            annot_template["colour"] = self.get_colors(attention[i] if isinstance(attention, list) else None)
            annot_template["attn"] = attention[i] if isinstance(attention, list) else attention
            annot_template["representation"] = representation[i] if isinstance(representation, list) else representation
            layer_template["annots"].append(annot_template)


        return layer_template

    def create_layer(self, boundaries, layer_num):

        layer_template = copy.deepcopy(self.layer_template)
        layer_template["name"] = f'layer {layer_num}'
        layer_template["order"] = layer_num

        for i in range(len(boundaries)-1):

            annot_template = copy.deepcopy(self.annot_template)
            annot_template["start"] = boundaries[i]
            annot_template["end"] = boundaries[i+1]
            annot_template["colour"] = self.get_colors(attention[i] if isinstance(attention, list) else None)
            annot_template["attn"] = attention[i] if isinstance(attention, list) else attention
            annot_template["representation"] = representation[i] if isinstance(representation, list) else representation
            layer_template["annots"].append(annot_template)

        return layer_template

    def __call__(self, filepath, duration, hierarchy):


        h_boundaries = hierarchy['boundaries']

        output_file = copy.deepcopy(self.json_template)
        output_file["duration"] = duration

        if self.is_inference:
            output_file["file"] =  filepath
        else:
            output_file["file"] =  os.path.join(self.prefix, *filepath.split('/')[-2:]) + f'.{self.postfix}'


        for layer_num in range(len(h_boundaries)):

            layer_template = copy.deepcopy(self.layer_template)
            layer_template["name"] = f'layer {layer_num}'
            layer_template["order"] = layer_num

            boundaries = h_boundaries[layer_num]
            for i in range(len(boundaries)-1):
                annot_template = copy.deepcopy(self.annot_template)
                annot_template["start"] = boundaries[i]
                annot_template["end"] = boundaries[i+1]
                annot_template["colour"] = self.get_colors(attn=None)
                annot_template["action"] = "N/A"
                layer_template["annots"].append(annot_template)

            output_file["layers"].append(layer_template)

        if not self.is_inference:
            filename = os.path.splitext(os.path.basename(filepath))[0]
            json.dump(output_file, open(os.path.join(self.json_dir, filename+'.json'), 'w'))

        return output_file


    def log_with_attn(self, filepath, duration, hierarchy):

        h_boundaries = hierarchy['boundaries']
        h_attention = hierarchy['attention']

        output_file = copy.deepcopy(self.json_template)
        output_file["duration"] = duration

        if self.is_inference:
            output_file["file"] =  filepath
        else:
            output_file["file"] =  os.path.join(self.prefix, *filepath.split('/')[-2:]) + f'.{self.postfix}'


        for layer_num in range(len(h_boundaries)):
            at = self.get_attn_norm(h_attention[layer_num])

            if layer_num == 0:
                bs = np.linspace(start=0.0, stop=len(at)*self.snippet_size, num=len(at)+1, endpoint=True)
            else:
                bs = h_boundaries[layer_num-1]

            layer = self.create_layer(bs, layer_num, at)
            output_file["layers"].append(layer)


        layer = self.create_layer(h_boundaries[layer_num], layer_num+1)
        output_file["layers"].append(layer)

        if not self.is_inference:
            filename = os.path.splitext(os.path.basename(filepath))[0]
            json.dump(output_file, open(os.path.join(self.json_dir, filename+'.json'), 'w'))

        return output_file


class VideoWriterFFMPEG():
    def __init__(self, path, fps):
        self.video_writer = FFmpegWriter(path, inputdict={'-r': str(fps)}, outputdict={'-r': str(fps)})

    def __call__(self, frame):
        self.video_writer.writeFrame(frame)

    def __del__(self):
        self.video_writer.close()


class VideoWriterInference(VideoWriterFFMPEG):
    def __init__(self, path, fps):
        super(VideoWriterInference, self).__init__(path, fps)


        # Create a white background image
        self.width = 1280
        self.height = 500

        # Define the dimensions of small images
        self.small_width = 100
        self.small_height = 100
        self.small_padding = 25
        self.small_padding_y = 10

        # Define the dimensions of large images
        self.large_width = 300
        self.large_height = 300
        self.large_padding = 95
        self.large_padding_y = 10


        # Define text parameters
        font_path = os.path.join(cv2.__path__[0],'qt','fonts','DejaVuSans.ttf')
        self.font = ImageFont.truetype(font_path, size=64)
        self.text_color = (255, 0, 0)  # Red text color

        # avg buffer
        self.buffer = deque(maxlen=10)


    def __call__(self, buffer, prediction, groundtruth, text):


        # small images
        small_images = list(map(lambda x: (x.squeeze(0).permute(1,2,0).cpu().numpy()+1.0)/2.0, buffer))

        # buffer image
        # buffer_avg = np.mean(small_images, axis=0) if small_images else np.zeros((self.large_width,self.large_height,3))

        # prediction image
        prediction = (prediction.squeeze(0).permute(1,2,0).detach().cpu().numpy()+1.0)/2.0 if prediction != None else np.zeros((self.large_width,self.large_height,3))

        # groundtruth
        groundtruth = (groundtruth.squeeze(0).permute(1,2,0).cpu().numpy()+1.0)/2.0

        # buffer
        self.buffer.append(groundtruth)
        buffer_gt = np.mean(list(self.buffer), axis=0) if self.buffer else groundtruth

        # large images
        large_images = [buffer_gt, prediction, groundtruth]


        image = Image.new("RGB", (self.width, self.height), (255, 255, 255))
        draw = ImageDraw.Draw(image)

        # Paste small images in the first row
        x = self.small_padding
        y = 0
        for small_image in small_images:
            small_image = Image.fromarray((small_image*255.0).astype(np.uint8)).resize((self.small_width, self.small_height))
            image.paste(small_image, (x, y))
            x += self.small_width + self.small_padding

        # Paste large images in the second row
        x = self.large_padding
        y += (self.small_height+ self.small_padding_y)
        for large_image in large_images:
            large_image = Image.fromarray((large_image*255.0).astype(np.uint8)).resize((self.large_width, self.large_height))
            image.paste(large_image, (x, y))
            x += self.large_width + self.large_padding


        # Add text to the image
        text_width, text_height = draw.textsize(text, font=self.font)
        x = (self.width - text_width) // 2
        y += (self.large_height + self.large_padding_y)
        draw.text((x, y), text, fill=self.text_color, font=self.font)

        # write frame to video
        super(VideoWriterInference, self).__call__(np.array(image))







import torch
import torch.nn as nn
import torch.nn.functional as F

class CNNEncoder(nn.Module):
    r"""
    A 4-layer CNN Encoder model used encode an image into a feature vector

    :param int feature_dim: the output feature dimension
    """

    def __init__(self, feature_dim):
        super(CNNEncoder, self).__init__()

        self.conv1 = nn.Conv2d(3, 32, kernel_size=3, padding=1)
        self.conv2 = nn.Conv2d(32, 64, kernel_size=3, padding=1)
        self.conv3 = nn.Conv2d(64, 256, kernel_size=3, padding=1)
        self.conv4 = nn.Conv2d(256, 1024 , kernel_size=3, padding=1)
        self.pool = nn.MaxPool2d(2, 2)
        self.fc1 = nn.Linear(1024 * 8 * 8, feature_dim)

    
    def forward(self, x):
        r"""
        The forward propagation function that takes input image and returns output vector

        :param torch.Tensor x: tensor of shape [1, 3, H, W]
        :returns:
            * (*torch.Tensor*): feature vector of shape [1, feature_dim]
            * (*None*): Used for compatibility to return attention in other models
        """

        x = F.relu(self.conv1(x))
        x = self.pool(x)
        x = F.relu(self.conv2(x))
        x = self.pool(x)
        x = F.relu(self.conv3(x))
        x = self.pool(x)
        x = F.relu(self.conv4(x))
        x = self.pool(x)
        x = x.view(-1, 1024 * 8 * 8)
        x = self.fc1(x)
        return x, None


class CNNDecoder(nn.Module):
    r"""
    A 4-layer CNN Decoder model used decode a feature vector back to an image

    :param int feature_dim: the input feature dimension
    """

    def __init__(self, feature_dim):
        super(CNNDecoder, self).__init__()

        self.fc1 = nn.Linear(feature_dim, 1024 * 8 * 8)
        self.conv1 = nn.Conv2d(1024, 256, kernel_size=3, padding=1)
        self.conv2 = nn.Conv2d(256, 64, kernel_size=3, padding=1)
        self.conv3 = nn.Conv2d(64, 32, kernel_size=3, padding=1)
        self.conv4 = nn.Conv2d(32, 3, kernel_size=3, padding=1)
    
    def forward(self, x):
        r"""
        The forward propagation function that takes a feature vector and returns an image

        :param torch.Tensor x: tensor of shape [1, feature_dim]
        :returns:
            (*torch.Tensor*): image tensor of shape [1, 3, H, W]
        """

        x = self.fc1(x)
        x = x.view(-1, 1024, 8, 8)
        x = F.interpolate(x, scale_factor=2)
        x = F.relu(self.conv1(x))
        x = F.interpolate(x, scale_factor=2)
        x = F.relu(self.conv2(x))
        x = F.interpolate(x, scale_factor=2)
        x = F.relu(self.conv3(x))
        x = F.interpolate(x, scale_factor=2)
        x = self.conv4(x)
        return x


class TemporalEncoding(nn.Module):
    r"""
    The temporal encoding module receives as an input a sequence of feature vectors [S, feature_dim] or a sequence of images [S, 3, H, W]
    and returns the a single summary feature vector [1, feature_dim].

    If the input is a sequence of images, it instantiates a user-defined encoder class to encode the images into a series of feature vectors

    :param int feature_dim: the input feature dimension
    :param int buffer_size: the maximum buffer size to create positional encoding
    :param float lr: the learning rate of this module
    :param int n_heads: the number of heads for the attention layer
    :param int num_layers: the number of transformer encoder layers
    :param torch.nn.Module encoder: the encoder class to be used. Default: None
    :param bool patch: patch the transformer model to retain attention information. Default: False
    """


    def __init__(self, feature_dim, buffer_size, lr, n_heads=4, num_layers=2, encoder=None, patch=False):
        super(TemporalEncoding, self).__init__()

        if encoder != None: self.encoder = encoder(feature_dim)
        self.cls_token = nn.Embedding(1, feature_dim, max_norm=1.0)
        self.pos_enc = nn.Embedding(buffer_size, feature_dim, max_norm=1.0)
        encoder_layer = nn.TransformerEncoderLayer(d_model=feature_dim, nhead=n_heads, dim_feedforward=feature_dim*2, activation=nn.ReLU(), dropout=0.0)
        norm = nn.LayerNorm(feature_dim, elementwise_affine=False)
        self.network = nn.TransformerEncoder(encoder_layer, num_layers=num_layers, norm=norm)
        if patch: self.patcher = Patcher(self.network, patch=True) # patch and hook transformer

        # optimizer
        self.optim= torch.optim.Adam(self.parameters(), lr=lr)

    def get_params(self):
        r"""
        Function to extract the parameters of this module

        :returns:
            (*List(torch.tensor)*): List of parameters

        """
        return list(self.parameters())

    def get_grads(self, loss):
        r"""
        Computes the gradients with torch.autograd.grad.

        Not used in the current implementation.

        :returns:
            (*List(torch.tensor)*): List of gradients

        """
        if hasattr(self, 'optim'):
            return torch.autograd.grad(loss, self.optim.param_groups[0]['params'] , retain_graph=True, allow_unused=True)

    def step_params(self):
        r"""
        Applies gradient step on the parameters of this module. Called by :py:class:`~streamer.optimizers.streamer_optimizer.StreamerOptimizer`.
        """
        if hasattr(self, 'optim'): 
            self.optim.step()

    def zero_params(self):
        r"""
        Zeros out the gradients of the parameters of this module. Called by :py:class:`~streamer.optimizers.streamer_optimizer.StreamerOptimizer`.
        """
        if hasattr(self, 'optim'): 
            self.optim.zero_grad()



    def forward(self, x):
        r"""
        Forward propagation function that receives a sequence of inputs and returns a single feature vector summarizing the sequence.

        :param torch.Tensor x: an input a sequence of feature vectors [S, feature_dim] or a sequence of images [S, 3, H, W]

        :returns:
            * (*torch.tensor*): The output feature vector representation of the input sequence
            * (*torch.tensor* or *None*): The output attention values of the encoder. Return None if encoder not defined or if using CNNEncoder
        """

        # encode input if needed
        attns = None
        if hasattr(self, "encoder"): 
            x, attns = self.encoder(x) #[S,D]

        # foward pass through transformer
        # x = F.normalize(x, p=2, dim=-1)
        x_pos = x.unsqueeze(1) + self.pos_enc(torch.arange(x.shape[0]).cuda()).unsqueeze(1) #[S,1,D]
        x_cls = torch.cat([self.cls_token(torch.tensor([0]).cuda()).unsqueeze(0), x_pos], dim=0) #[S+1,1,D]
        representation = self.network(x_cls)[0]

        return representation, attns


class HierarchicalPrediction(nn.Module):
    r"""
    The hierarchical prediction modules receives as an input a sequence of feature vectors [S, feature_dim] where S is the number of layers
    and returns the a single prediction feature vector [1, feature_dim] or [1, 3, H, W] if a decoder class is provided.

    :param int feature_dim: the input feature dimension
    :param int max_layers: the maximum number of layers
    :param float lr: the learning rate of this module
    :param int layer_num: the layer number where this class is instantiated
    :param int n_heads: the number of heads for the attention layer
    :param int num_layers: the number of transformer encoder layers
    :param torch.nn.Module decoder: the decoder class to be used. Default: None
    :param bool patch: patch the transformer model to retain attention information. Default: False
    """

    def __init__(self, feature_dim, max_layers, lr, layer_num, n_heads=4, num_layers=2, decoder=None, patch=False):
        super(HierarchicalPrediction, self).__init__()

        self.layer_num = layer_num

        self.pos_enc = nn.Embedding(max_layers, feature_dim, max_norm=1.0)
        encoder_layer = nn.TransformerEncoderLayer(d_model=feature_dim, nhead=4, dim_feedforward=feature_dim*2, activation=nn.ReLU(), dropout=0.0)
        norm = nn.LayerNorm(feature_dim, elementwise_affine=False)
        self.network = nn.TransformerEncoder(encoder_layer, num_layers=num_layers, norm=norm)
        if decoder!= None: self.decoder = decoder(feature_dim)
        if patch: self.patcher = Patcher(self.network, patch=True) # patch and hook transformer

        # optimizer
        self.optim= torch.optim.Adam(self.parameters(), lr=lr)

    def get_params(self):
        r"""
        Function to extract the parameters of this module

        :returns:
            (*List(torch.tensor)*): List of parameters

        """
        return list(self.parameters())

    def get_grads(self, loss):
        r"""
        Computes the gradients with torch.autograd.grad.

        Not used in the current implementation.

        :returns:
            (*List(torch.tensor)*): List of gradients

        """
        if hasattr(self, 'optim'):
            return torch.autograd.grad(loss, self.optim.param_groups[0]['params'], retain_graph=True, allow_unused=True)

    def step_params(self):
        r"""
        Applies gradient step on the parameters of this module. Called by :py:class:`~streamer.optimizers.streamer_optimizer.StreamerOptimizer`.
        """
        if hasattr(self, 'optim'): 
            self.optim.step()

    def zero_params(self):
        r"""
        Zeros out the gradients of the parameters of this module. Called by :py:class:`~streamer.optimizers.streamer_optimizer.StreamerOptimizer`.
        """
        if hasattr(self, 'optim'): 
            self.optim.zero_grad()

    def forward(self, x):
        r"""
        Forward propagation function that receives a sequence of inputs and returns a single feature vector (or a single decoded image) predicting the next input.

        :param torch.Tensor x: an input a sequence of feature vectors [S, feature_dim]

        :returns:
            (*torch.tensor*): The output prediction feature vector [1, feature_dim] or decoded image [1, 3, H, W]
        """

        reps_pos = x.unsqueeze(1) + self.pos_enc(torch.arange(x.shape[0]).cuda()).unsqueeze(1)
        context = self.network(reps_pos)[self.layer_num]
        # context = F.normalize(context, p=2, dim=-1) 
        if hasattr(self, "decoder"): 
            context = self.decoder(context.unsqueeze(0))
            # context = F.normalize(context, p=2, dim=1) 

        return context


if __name__ == "__main__":

    # === Temporal Encoding === #

    encoding = TemporalEncoding(1024, 20, 1e-4, 4, 2, CNNEncoder).cuda()
    input = torch.randn(5,3,128,128).cuda()
    rep, attn = encoding(input)
    print(rep.shape)


    # === Hierarchical Prediction === #

    predictor = HierarchicalPrediction(1024, 3, 1e-4, 1, 4, 2, CNNDecoder).cuda()
    input = torch.randn(3,1024).cuda()
    context = predictor(input)
    print(context.shape)








from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as f:
  long_description = f.read()

setup(
  name="streamer-torch",
  version="0.0.1",
  build="0",
  author="Ramy Mounir",
  url="https://ramymounir.com/docs/streamer/",
  description=r"""Official implementation of STREAMER, a self-supervised hierarchical event segmentation and representation learning""",
  long_description=long_description,
  long_description_content_type="text/markdown",
  packages=find_packages(),
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: Free for non-commercial use",
  ],
  python_requires='>=3.10',
  install_requires=[
      'torch>=2.0.0',
      'ddpw>=5.1.0',
      'opencv-python-headless',
      'tqdm',
      'scikit-video',
      'matplotlib',
      'torchvision',
      'tensorboard',
      'pandas',
      'moviepy',
      'librosa',
      'pillow',
      ]
)



import torch, os
import torch.nn as nn
import torch.nn.functional as F
from streamer.models.layer import StreamerLayer, StreamerLayerArguments
from dataclasses import dataclass, asdict

@dataclass
class StreamerModelArguments():

    log_base_every: int = 1000
    r"""Tensorboard log of the base layer every 'log_base_every'"""

    main: bool = True
    r"""Main process/gpu or not"""

    max_layers: int = 3
    r"""The maximum number of layers to stack"""

    feature_dim: int = 1024
    r"""Feature dimension of the model embeddings"""

    evolve_every: int = 50000
    r"""Create/stack a new layer every 'evolve_every' """

    buffer_size: int = 10
    r"""Maximum input buffer size to be used"""

    force_fixed_buffer: bool = False
    r"""Force the buffer to be fixed (not replacing inputs) by triggering a boundary when buffer is full"""

    loss_threshold: float = 0.25
    r"""Loss threshold value. Not used in average demarcation mode"""

    lr: float = 1e-4
    r"""Learning rate to be used in all modules"""

    init_layers: int = 1
    r"""How many layers to initialize before training"""

    init_ckpt: str = ''
    r"""the path of the pretrained weights, if any"""

    ckpt_dir: str = ''
    r"""the path to save weights"""

    snippet_size: float = 0.5
    r"""Snippet size of input video (seconds/image). Typically 0.5 seconds per image"""

    demarcation_mode: str = 'average'
    r"""Demarcation mode used to detect boundaries"""

    distance_mode: str = 'distance'
    r"""Distance mode for loss calculation"""

    force_base_dist: bool = True
    r"""Force the lowest layer to use MSE instead of Cosine Similarity"""

    window_size: int = 50
    r"""Window size for average demarcation mode"""

    modifier_type: str = 'multiply'
    r"""Modifier type to apply to average demarcation mode ['multiply', 'add']"""

    modifier: float = 1.0
    r"""Modifier to apply to avrrage demarcation mode"""

    fp_up: bool = True
    r"""Allow bottom up inference"""

    fp_down: bool = True
    r"""Allow top down inference"""

    l1_loss_weight: float = 100
    r"""L1 loss weight. Not used in the current loss implementation"""


    @staticmethod
    def from_args(args):
        return StreamerModelArguments(
                init_ckpt=args.init_ckpt,
                init_layers=args.init_layers,
                ckpt_dir=args.ckpt_dir,
                snippet_size=args.snippet_size,
                log_base_every=args.log_base_every,
                main=args.main,
                max_layers=args.max_layers,
                feature_dim=args.feature_dim,
                evolve_every=args.evolve_every,
                buffer_size=args.buffer_size,
                force_fixed_buffer=args.force_fixed_buffer,
                loss_threshold=args.loss_threshold,
                l1_loss_weight=args.l1_loss_weight,
                lr=args.lr,
                demarcation_mode=args.demarcation_mode,
                distance_mode=args.distance_mode,
                force_base_dist=args.force_base_dist,
                window_size=args.window_size,
                modifier_type=args.modifier_type,
                modifier=args.modifier,
                fp_up=args.fp_up,
                fp_down=args.fp_down
        )


    @staticmethod
    def from_ckpt_args(args):
        return StreamerModelArguments(
                init_ckpt=args.init_ckpt,
                init_layers=args.init_layers,
                ckpt_dir=args.ckpt_dir,
                snippet_size=args.snippet_size,
                log_base_every=args.log_base_every,
                main=args.main,
                max_layers=args.max_layers,
                feature_dim=args.feature_dim,
                evolve_every=args.evolve_every,
                buffer_size=args.buffer_size,
                force_fixed_buffer=args.force_fixed_buffer,
                loss_threshold=args.loss_threshold,
                l1_loss_weight=args.l1_loss_weight,
                lr=args.lr,
                demarcation_mode=args.demarcation_mode,
                distance_mode=args.distance_mode,
                force_base_dist=args.force_base_dist,
                window_size=args.window_size,
                modifier_type=args.modifier_type,
                modifier=args.modifier,
                fp_up=args.fp_up,
                fp_down=args.fp_down
        )



class StreamerModel(nn.Module):
    r"""
    The implementation of the STREAMER model for training. This class initializes the first layer(s) of streamer, saves/loads weights, etc.

    :param StreamerModelArguments args: The arguments passed to Streamer Model
    :param Logger logger: The tensorboard logger class
    :param torch.nn.Module encoder: The encoder model (e.g., :py:class:`~streamer.models.networks.CNNEncoder`)
    :param torch.nn.Module decoder: The decoder model (e.g., :py:class:`~streamer.models.networks.CNNDecoder`)
    """

    def __init__(self, args:StreamerModelArguments, logger=None, encoder=None, decoder=None):
        super(StreamerModel, self).__init__()
        self.args = args
        self.logger = logger

        self.streamer = None
        self.models_saved = 0

        # counters
        self.base_counter = 0
        self.__initialize(args.init_ckpt, args.init_layers, encoder=encoder, decoder=decoder)



    def init_layer(self, count=1, encoder=None, decoder=None):
        r"""
        Initializes the Streamer layer(s) and passes the enocder/decoder to the first layer.
        
        :param int count: How many layers to create 
        :param torch.nn.Module encoder: The encoder model (e.g., :py:class:`~streamer.models.networks.CNNEncoder`)
        :param torch.nn.Module decoder: The decoder model (e.g., :py:class:`~streamer.models.networks.CNNDecoder`)
        """

        # Arguments
        self.args.reps_fn = self.getReps
        streamerLayerArgs = StreamerLayerArguments.from_model_args(self.args)

        self.streamer = StreamerLayer(
                args = self.args,
                layer_num = 0, 
                init_count = count,
                encoder=encoder,
                decoder=decoder,
                logger=self.logger
                )


    def __initialize(self, ckpt='', count=1, encoder=None, decoder=None):
        r"""
        Initializes the Streamer layer(s) with checkpoint if available.
        
        :param str ckpt: pretrained weights location
        :param int count: How many layers to create 
        :param torch.nn.Module encoder: The encoder model (e.g., :py:class:`~streamer.models.networks.CNNEncoder`)
        :param torch.nn.Module decoder: The decoder model (e.g., :py:class:`~streamer.models.networks.CNNDecoder`)
        """

        if ckpt=='':
            self.init_layer(count=count, encoder=encoder, decoder=decoder)

        else:
            ckpt= torch.load(ckpt, map_location='cpu')
            self.init_layer(count=ckpt['num_layers'], encoder=encoder, decoder=decoder)
            self.streamer.load_state_dict(ckpt['weights'])
            

    def forward(self, x):
        r"""
        Forward propagation function that calls the :py:meth:`~streamer.models.layer.StreamerLayer.forward` function of the first :py:class:`~streamer.models.layer.StreamerLayer`.
        
        :param torch.Tensor x: the input image [1, 3, H, W]
        """

        # log base signal
        if self.logger != None: 
            self.logger.model(self.streamer.context, 
                              x, 
                              self.streamer.attn_img, 
                              self.streamer.attns)

        # main forward
        self.streamer(x, base_counter=self.base_counter)
        self.base_counter += 1


    def getReps(self, layer_num):
        r"""
        Aggregates the representations from all the layers.

        :param int layer_num: the index of the calling layer

        :returns:
            (*torch.Tensor*): concatenated representations from all layers [L, feature_dim]

        """

        def check_fp(layer, curr_layer):
            down, up = (layer < curr_layer), (layer > curr_layer)

            if not (up or down): return True
            if up and self.args.fp_up: return True
            if down and self.args.fp_down: return True
            return False

        curr_layer_num = 0
        ll = self.streamer
        reps = [ll.representation] if check_fp(layer_num, curr_layer_num) else [ll.representation.detach()]

        while ll.above != None and ll.above.representation != None:
            curr_layer_num += 1
            ll = ll.above
            if check_fp(layer_num, curr_layer_num): reps.append(ll.representation)
            else: reps.append(ll.representation.detach())

        return torch.cat(reps, dim=0)

    def extract_rep(self):
        r"""
        Extract representation function used for logging json hierarchy. 
        Calls recursive function :py:meth:`~streamer.models.layer.StreamerLayer.extract_rep` on the :py:class:`~streamer.models.layer.StreamerLayer` class

        :returns:
            (*dict*): hierarchy represented as boundaries of every layer

        """

        if self.streamer == None:
            return {}

        hierarchy = dict(boundaries = {})
        self.streamer.extract_rep(hierarchy)

        # close open boundaries
        for layer_name, layer_bounds in hierarchy['boundaries'].items():
            if layer_name == 0: continue
            if layer_bounds[-1] != hierarchy['boundaries'][0][-1]:
                layer_bounds.append(hierarchy['boundaries'][0][-1])

        return hierarchy

    def reset_model(self):
        r"""
        Resets the whole streamer model for a new video.
        Calls recursive function :py:meth:`~streamer.models.layer.StreamerLayer.reset_layer` on the :py:class:`~streamer.models.layer.StreamerLayer` class

        """
        if self.streamer == None:
            return 

        self.streamer.reset_layer()
        self.base_counter = 0

    def get_num_layers(self):
        r"""
        Get the total number of layers.
        Calls recursive function :py:meth:`~streamer.models.layer.StreamerLayer.get_num_layers` on the :py:class:`~streamer.models.layer.StreamerLayer` class
        """
        if self.streamer == None:
            return 0

        return self.streamer.get_num_layers(0)

    def optimize_model(self):
        r"""
        Optimizes the whole streamer model (gradient step). 
        Calls recursive function :py:meth:`~streamer.models.layer.StreamerLayer.optimize_layer` on the :py:class:`~streamer.models.layer.StreamerLayer` class
        """
        if self.streamer != None:
            self.streamer.optimize_layer()
        

    def save_model(self):
        r"""
        Saves the model weights to :py:class:`~StreamerModelArguments.ckpt_dir`
        """

        if self.streamer == None or self.args.main == False:
            return 

        ckpt = dict(
                model_args = asdict(self.args),
                model_modality = self.modality,
                model_snippet_size = self.args.snippet_size,
                num_layers = self.get_num_layers(), 
                weights = self.streamer.state_dict())

        torch.save(ckpt, os.path.join(self.args.ckpt_dir, f'model_{str(self.models_saved).zfill(3)}.pth'))
        self.models_saved += 1




import torch.nn as nn

class StreamerLoss(nn.Module):
    r"""
    Streamer loss function is applied to minimize the distance between the prediction and actual input.
    Implemented for Cosine Similarity and Euclidean distance.
    Use only Cosine Similarity as it backpropagates bounded loss values across layers.

    :param str dist_mode: The distance mode: ['similarity', 'distance']
    """

    def __init__(self, dist_mode):
        super(StreamerLoss, self).__init__()
        self.dist_mode = dist_mode
        self.criterion = nn.CosineSimilarity() if dist_mode=='similarity' else nn.MSELoss()


    def forward(self, context, groundtruth):
        r"""
        Calculates the loss value from context and groundtruth

        :param torch.Tensor context: The prediction tensor. shape=[1,3,H,W] or [1,feature_dim]
        :param torch.Tensor groundtruth: The groundtruth tensor. shape=[1,3,H,W] or [1,feature_dim]
        :returns: 
            * (*dict*): dictionary of all the losses
            * (*float*): the loss value to be used for demarcation (i.e., input to the :py:meth:`~streamer.core.demarcation.EventDemarcation.__call__`)
        """

        loss = self.criterion(context, groundtruth).mean()
        losses_dict = dict(l2_loss = loss)
        return losses_dict, loss

    def get_params(self): return []
    def step_params(self): pass
    def zero_params(self): pass

    def summarize_loss(self, loss_dict):
        r"""
        Converts losses_dict from the output of :py:meth:`.forward` to loss value respecting the distance mode.
        Called by the optimizer

        :param dict losses_dict: The full losses dictionary
        :returns:
            (*float*): The loss value to be minimized
        """
        loss_sign = -1 if self.dist_mode == 'similarity' else 1
        return loss_sign * sum(list(loss_dict.values()))






import torch
import torch.nn as nn
import torch.nn.functional as F


class CounterDetector:
    def __init__(self, count):
        self.count = count
        self.counter = 0

    def __call__(self):

        if self.counter > self.count:
            return True
        else:
            self.counter += 1

        return False

class CycleDetector:
    def __init__(self, count):
        self.count = count
        self.counter = 1

    def __call__(self):

        if self.counter == self.count:
            self.counter = 1
            return True
        else:
            self.counter += 1
            return False


class PlateauDetector:
    def __init__(self, window_size=10, threshold=1e-6):
        self.window_size = window_size
        self.threshold = threshold
        self.values = []
        self.mean = None

    def __call__(self, value):
        self.values.append(value)

        if len(self.values) <= self.window_size:
            return False

        self.values.pop(0)
        mean = torch.tensor(self.values).mean()

        if (value>=0.5) and (self.mean is not None) and (math.isclose(mean, self.mean, rel_tol=self.threshold)):
            return True

        self.mean = mean

        return False


class Patcher():

    def __init__(self, model, patch=True):
        self.reset()
        self.num_layers = len(model.layers)

        for l in model.layers:
            if patch: self.patch_layer(l.self_attn)
            l.self_attn.register_forward_hook(self.save_output)

    def patch_layer(self, m):

        forward_orig = m.forward

        def wrap(*args, **kwargs):
            kwargs["need_weights"] = True
            kwargs["average_attn_weights"] = True

            return forward_orig(*args, **kwargs)

        m.forward = wrap

    def save_output(self, m, m_in, m_out):
        if len(self.outputs) == self.num_layers:
            self.reset()

        self.outputs.append( m_out[1].mean(dim=0) )

    def calculate_attn(self, index=0):
        if self.attention_calculated: return None

        self.outputs.reverse()

        self.final_attention = self.outputs[0]
        for attn in self.outputs[1:]:
            self.final_attention = self.final_attention@attn

        self.attention_calculated = True

        attn_lst = self.final_attention[index][1:]
        attn_lst /= sum(attn_lst)
        return attn_lst.tolist()

    def reset(self):
        self.outputs = []
        self.final_attention= None
        self.attention_calculated = False





import os, json, cv2, random
import os.path as osp
import numpy as np
from glob import glob
from torch.utils.data import Dataset
from torchvision import transforms as T
from dataclasses import dataclass


@dataclass
class EpicDatasetArguments():
    r"""
    Arguments for Epic-Kitchens dataset
    """

    world_size: int
    r"""Number of gpus to distribute the dataset"""

    global_rank: int
    r"""The rank of the running device"""

    dataset: str
    r"""The dataset name"""

    frame_size: list
    r"""The frame size to which the images will be resized"""

    percentage: int
    r"""Percentage of the dataset to run"""

    split: str = 'train'
    r"""The split of data to use. Choices: ['train', 'test']"""

    @staticmethod
    def from_args(args):
        return EpicDatasetArguments(
                world_size=args.world_size, 
                global_rank=args.global_rank, 
                dataset=args.dataset, 
                frame_size=args.frame_size,
                percentage=args.dataset_percent,
                split=args.dataset_split
        )


class EpicDataset(Dataset):
    r"""
    The epic-kitchens dataset that iterates over the whole dataset and returns the frames one by one in a streaming fashion.

    :param EpicDatasetArguments args: The parameters used for the Epic-Kitchens dataset
    """


    def __init__(self, args: EpicDatasetArguments):

        vids_map_path = os.path.join(args.dataset, 'map.json')
        if not os.path.exists(vids_map_path):
            quit('mapping file does not exist')

        with open(vids_map_path, 'r') as f:
            vid_paths_dict_full = json.load(f)
            self.snippet_size = vid_paths_dict_full['meta']['snippet_size']
            self.vids_paths = vid_paths_dict_full['files']

        # filter only split videos
        if osp.exists(osp.join(args.dataset, 'groundtruth')):
            split_path = osp.join(args.dataset, 'groundtruth', args.split, '**/*.json')
            train_vids = [osp.splitext(osp.basename(f))[0] for f in glob(split_path)]
            self.vids_paths = dict(filter(lambda x: osp.splitext(osp.basename(x[0]))[0] in train_vids, self.vids_paths.items()))

        # sample a percentage of files
        self.vids_paths = dict(random.sample(self.vids_paths.items(), int(len(self.vids_paths)*args.percentage/100.0)))
        print(f'processing {len(self.vids_paths)} for the {args.split} split')

        self.min_sum, self.gpu_train_paths = self.split_dict_greedy(
                {k:[v, v] for k,v in self.vids_paths.items()}, 
                args.world_size,
                args.global_rank)

        self.prepare_data(make_equal=(args.split=='train'))
        self.len = len(self.data)

        self.transform = T.Compose([
                    T.ToTensor(),
                    T.Resize(args.frame_size, antialias=True),
                ])

    def prepare_data(self, make_equal=True):

        to_go = self.min_sum

        self.data = []
        for video in self.gpu_train_paths:
            img_paths = np.array(sorted(glob(f'{video}/*.jpg')))
            duration = self.vids_paths[video] * self.snippet_size

            for i in range(self.vids_paths[video]):

                self.data.append([video, 
                                  duration, 
                                  i, 
                                  i*self.snippet_size, 
                                  img_paths[int(i*(len(img_paths)//self.vids_paths[video]))], 
                                  False])
                to_go -= 1
                if to_go == 0 and make_equal: break

            self.data[-1][-1] = True
            if to_go == 0 and make_equal: break

    # Split array into chunks using greedy. Courtesy of Bing AI
    def split_dict_greedy(self, dct, n, r):
        # Get the keys and values from the dictionary and sort them by values in descending order
        keys_values = list(zip(dct.keys(), dct.values()))
        keys_values.sort(key=lambda x: x[1][1], reverse=True)
        keys, lst = zip(*keys_values)

        # Convert keys and lst into lists
        keys = list(keys)
        lst = list(lst)

        # Initialize empty chunks and sums
        chunks = [[] for _ in range(n)]
        sums = [0] * n

        # Loop through the list and assign each element to the chunk with the smallest sum
        for i in range(len(lst)):
            min_index = sums.index(min(sums))
            chunks[min_index].append(keys[i])
            sums[min_index] += lst[i][1]

        return min(sums), chunks[r]


    def __getitem__(self, index):
        r"""
        Iterates over the dataset in a streaming fashion and retrieves one frame at a time

        :param int index: The index of the item in the dataset to retrieve
        :returns:
            * (*torch.tensor*): the frame in tensor format
            * (*List*): Information about the frame [Video, duration, index, time, last_frame]
        """

        info = self.data[index]
        frames = self.transform(cv2.cvtColor(cv2.imread(info[4]), cv2.COLOR_BGR2RGB))*2.0-1.0
        return frames, info

    def __len__(self):
        return self.len
